#include "stdafx.h"
#include "LogViewList.h"
#include "StringHelper.h"

#define BUFSIZE			1023

#define SEPARATOR1		" ; "
#define SEPARATOR2		"\t"
#define SEPARATOR3		"\";\"" //";"
#define SEPARATOR4		"\",\""	//","
#define SEPARATOR5		"\" \"" //" "
#define SEPARATOR6		";"
#define SEPARATOR7		","
#define SEPARATOR8		" "

#define METALINE		"#"
#define HEADERFIELD		"#Fields: "
#define HEADERSOFTWARE	"#Software: "
#define HEADERDATE		"#Date: "

#define ID_LIST			31978

struct PARAMSORT
    {
        PARAMSORT(CListCtrl* pList, int columnIndex, bool ascending)
            :m_pList(pList)
            ,m_ColumnIndex(columnIndex)
            ,m_Ascending(ascending)
        {}

        CListCtrl* m_pList;
        int  m_ColumnIndex;
        bool m_Ascending;
    };


IMPLEMENT_DYNCREATE(CLogViewList, CView)

BEGIN_MESSAGE_MAP(CLogViewList, CView)
	ON_WM_CREATE()
	ON_WM_SIZE()
	ON_NOTIFY(HDN_ITEMCLICK, 0, OnItemclickListSort)
END_MESSAGE_MAP()

CLogViewList::CLogViewList(void):AbortBinding(false),fn(""),CurrentPosition(0),ColumnCount(0),HeaderName(""),HeaderDate(""),filtered_count(0)
{
}

CLogViewList::~CLogViewList(void)
{
}

BOOL CLogViewList::PreCreateWindow(CREATESTRUCT& cs)
{
	// Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CView::PreCreateWindow(cs);
}

// CsplitterbasicView drawing

void CLogViewList::OnDraw(CDC* /*pDC*/)
{
	//CsplitterbasicDoc* pDoc = GetDocument();
	//ASSERT_VALID(pDoc);
	//if (!pDoc)
	//	return;

	// add draw code for native data here
}


// CsplitterbasicView printing

BOOL CLogViewList::OnPreparePrinting(CPrintInfo* pInfo)
{
	// default preparation
	return DoPreparePrinting(pInfo);
}

void CLogViewList::OnBeginPrinting(CDC* pDC, CPrintInfo* pInfo)
{
	// TODO: add extra initialization before printing

	//source: http://www.codersource.net/2010/01/30/mfc-print-tutorial/
	//int nPageHeight = pDC->GetDeviceCaps(VERTRES);
	//int nDocLength = GetDocument()->DocLength();
	//int nMaxPage = max(1, (nDocLength+nPageHeight-1)/nPageHeight);
	//pInfo->SetMaxPage(nMaxPage);
}

void CLogViewList::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add cleanup after printing
}

void CLogViewList::PrintLog()
{ 
	this->OnFilePrintPreview();
	//this->OnFilePrint();
}

bool CLogViewList::ReadLog(LPCTSTR FileName, LONGLONG Start)
{
	try{
		AbortBinding = true;

		fn = FileName;
		CString line("");
		char bufread[BUFSIZE+1]; //+1 for termination
		//read log
		CFile flog;
		if (!flog.Open(fn,CFile::modeRead|CFile::shareDenyNone,NULL)){
			return false;
		}else{
			CStringList lines;
			UINT size(0);
			int eollength = Settings.General.EOL.GetLength();

			//read first chunk for headers or move to certain position
			if(Start>0){
				flog.Seek(Start,CFile::begin);
			}else{
				bool firstline = true;
				size = flog.Read(bufread,BUFSIZE);
				if(size>0){
					bufread[size] = '\0';
					POSITION pos;
					CStringList Headers;
					//detect EOL
					Settings.General.EOL = DetectEOL(bufread);
					eollength = Settings.General.EOL.GetLength();

					CStringHelper::Split(Headers,bufread,Settings.General.EOL);
					pos = Headers.GetHeadPosition();
					while(pos!=NULL){
						line = Headers.GetNext(pos);
						if (line.Left(strlen(METALINE))==METALINE){
							SetHeaders(line); //parse headers (columns,...)
							firstline = false;
						}else{
							//first log entry are column names
							if(firstline && Settings.General.FirstLineHeader){
								//if(Settings.General.Separator=="")
								//	Settings.General.Separator = DetectSeparator(line);  //don't do this: unable to manually select something else
								SetColumns(line,true);
								firstline = false;
							}else{
								//log entries
								lines.AddTail(line);
							}
						}
					}
					//add CRLF if that were the last 2 chars (removed by Split())
					if(size>eollength){
						CString endline("");
						for(int i = size-eollength; i < size; i++){
							endline += bufread[i];
						}
						if(endline == Settings.General.EOL){
							line += Settings.General.EOL;
						}
					}

					//add columns by manually set separator if none defined in log file
					if(ColumnCount==0 && Settings.General.Separator.GetLength()>0){
						POSITION pos = lines.GetHeadPosition();
						if(pos!=NULL){
							SetColumns(lines.GetNext(pos),false);
						}
					}

					//add a single column if there are none
					if(ColumnCount==0){
						CString s;
						s.LoadString(IDS_APP_LOG_ENTRIES);
						loglist.InsertColumn(0,s,LVCFMT_LEFT,600);
						ColumnCount++;
					}

					//remove last read incomplete line, will be added later
					if(!lines.IsEmpty())
						lines.RemoveTail();

					//add log entries
					AddLines(lines);
				}
			}
					
			//read rest of file
			size = flog.Read(bufread,BUFSIZE);
			while (size>0){
				bufread[size] = '\0'; //make sure it is null terminated
				CStringHelper::Split(lines,line + CString(bufread),Settings.General.EOL);
				line = lines.GetTail();

				//add CRLF if that were the last 2 chars (removed by Split())
				if(size>eollength){
					CString endline("");
					for(int i = size-eollength; i < size; i++){
						endline += bufread[i];
					}
					if(endline == Settings.General.EOL){
						line += Settings.General.EOL;
					}
				}
				if(!lines.IsEmpty())
					lines.RemoveTail();
				AddLines(lines);
				size = flog.Read(bufread,BUFSIZE);
			}

			//last line
			if(line!=""){
				//remove last CRLF as nothing more will be read
				if(line.Right(Settings.General.EOL.GetLength())==Settings.General.EOL){
					line = line.Left(line.GetLength()-Settings.General.EOL.GetLength());
				}
				AddLine(line);
			}

			//remember position and close
			CurrentPosition = (ULONGLONG)flog.GetPosition();
			flog.Close();
		}
		AbortBinding = false;


		return true;
	
	}catch(...){
		AbortBinding = false;
		return false;
	}
}

bool CLogViewList::ReadLogAppend()
{
	if(fn==""){
		return false;
	}else{
		return ReadLog(fn,CurrentPosition);
	}
}

inline void CLogViewList::Clear()
{
	FieldList.RemoveAll();
	loglist.DeleteAllItems();			//delete all rows
	while(loglist.DeleteColumn(0)!=0);	//delete all columns
	AbortBinding = false;
	//fn = ""; don't clear this: see RefreshLog();
	CurrentPosition = 0;
	ColumnCount = 0;
	HeaderName = "";
	HeaderDate = "";
	//Separator = "";
	filtered_count = 0;
}

void CLogViewList::CloseLog()
{
	Clear();
	fn="";
}

bool CLogViewList::HasFile()
{
	return fn!="";
}

void CLogViewList::RefreshLog()
{
	Clear();
	ReadLog(fn);
}

void CLogViewList::SetHeaders(CString& Header)
{
	try{

	//#Fields:
	if (Header.Left(strlen(HEADERFIELD))==HEADERFIELD && FieldList.GetSize()==0){
		Settings.General.Separator = DetectSeparator(Header);
		SetColumns(Header.Mid(strlen(HEADERFIELD)),true);
	}else{
		if(Header.Left(strlen(HEADERSOFTWARE))==HEADERSOFTWARE){
			HeaderName = Header.Mid(strlen(HEADERSOFTWARE));
		}else{
			if(Header.Left(strlen(HEADERDATE))==HEADERDATE){
				HeaderDate = Header.Mid(strlen(HEADERDATE));
			}
		}
	}

	}catch(...){
		//ignore error
	}
}

CString CLogViewList::DetectSeparator(CString& Line)
{
	if(Line.Find(SEPARATOR1,0)!=-1){
		return SEPARATOR1;
	}else{
		if(Line.Find(SEPARATOR2,0)!=-1){
			return SEPARATOR2;
		}else{
			if(Line.Find(SEPARATOR3,0)!=-1){
				return SEPARATOR3;
			}else{
				if(Line.Find(SEPARATOR4,0)!=-1){
					return SEPARATOR4;
				}else{
					if(Line.Find(SEPARATOR5,0)!=-1){
						return SEPARATOR5;
					}else{
						if(Line.Find(SEPARATOR6,0)!=-1){
							return SEPARATOR6;
						}else{
							if(Line.Find(SEPARATOR7,0)!=-1){
								return SEPARATOR7;
							}else{
								if(Line.Find(SEPARATOR8,0)!=-1){
									return SEPARATOR8;
								}else{
									return ""; //if no header separator found, use complete line as header
								}
							}
						}
					}
				}
			}
		}
	}
}

CString CLogViewList::DetectEOL(CString Text)
{
	if(Text.Find("\r\n",0)!=-1){ //DOS
		return "\r\n";
	}else{
		if(Text.Find("\n",0)!=-1){ //Unix
			return "\n";
		}else{
			if(Text.Find("\r",0)!=-1){ //Old Mac
				return "\r";
			}else{
				return "\r\n"; //if no EOL separator found, use CRLF
			}
		}
	}
}

void CLogViewList::SetColumns(CString& Fields, bool UseFieldNameAsColumnHeading)
{
	CStringList Flds;
	CStringHelper::Split(Flds,Fields,Settings.General.Separator);
	if(Settings.General.Trim){
		POSITION pos = Flds.GetHeadPosition();
		while(pos!=NULL){
			CString& fld = Flds.GetNext(pos);
			fld = fld.Trim().Trim('"');
		}
	}
	SetColumns(Flds,UseFieldNameAsColumnHeading);
}

void CLogViewList::SetColumns(CStringList &Fields, bool UseFieldNameAsColumnHeading)
{
	ColumnCount = 0;
	FieldList.RemoveAll();
	POSITION pos;
	CString fld;
	pos = Fields.GetTailPosition();
	while(pos!=NULL){
		fld = Fields.GetPrev(pos);
		loglist.InsertColumn(0,UseFieldNameAsColumnHeading?fld:"",LVCFMT_LEFT,CalculateColumnWidth(fld,70,500));
		ColumnCount++;
	}

	//internal headerlist
	if(UseFieldNameAsColumnHeading){
		FieldList.SetSize(ColumnCount);
		pos = Fields.GetHeadPosition();
		int i=0;
		while(pos!=NULL){
			FieldList.SetAt(i,Fields.GetNext(pos));
			i++;
		}
	}
}

int CLogViewList::CalculateColumnWidth(CString& Fld, int MinimumLength, int MaximumLength)
{
	const int FONTWIDTH_ = 10;
	int Length = Fld.GetLength();
	if(Length>MaximumLength/FONTWIDTH_){
		return MaximumLength;
	}else{
		if(Length<MinimumLength/FONTWIDTH_){
			return MinimumLength;
		}else{
			return Length*FONTWIDTH_;
		}
	}
}

void CLogViewList::AddLines(CStringList &lines)
{
	POSITION pos = lines.GetHeadPosition();
	while(pos!=NULL){
		AddLine(lines.GetNext(pos));
	}
	lines.RemoveAll();
}

bool CLogViewList::AddLine(CString& Line)
{
	try{
		if(Line.Left(strlen(METALINE))==METALINE){
			return false;
		}else{
			int lnumber = loglist.GetItemCount();
			if(Settings.IsAllowedByFilters(Line) && !(Settings.Filter.MaxEntries > 0 && lnumber == Settings.Filter.MaxEntries) ){
				int i(0);
				POSITION pos;
				CStringList FieldItems;
				CStringHelper::Split(FieldItems,Line,Settings.General.Separator,ColumnCount);
				lnumber = loglist.InsertItem(lnumber+1,Line);
				if(lnumber==-1){
					return false;
				}else{
					pos = FieldItems.GetHeadPosition();
					if(Settings.General.Trim){
						while (pos!=NULL){
							CString Item;
							Item = FieldItems.GetNext(pos).Trim().Trim('"');
							loglist.SetItem(lnumber,i,LVIF_TEXT,Item,0,0,0,0);
							i++;
						}
					}else{
						while (pos!=NULL){
							CString Item;
							Item = FieldItems.GetNext(pos);
							loglist.SetItem(lnumber,i,LVIF_TEXT,Item,0,0,0,0);
							i++;
						}
					}
					return true;
				}
			}else{
				filtered_count++;
				return false;
			}
		}
	
	}catch(...){
		return false;
	}
}

void CLogViewList::BindDetailToList()
{
	CString Row = GetCurrentEntry();
	if(Row.GetLength()>0){
		pDetails->SetDetailView(Row,Settings.General.Separator,ColumnCount);
	}
}

CString CLogViewList::GetCurrentEntry()
{
	CString Row("");
	if(!AbortBinding){
		static const int bufsize = 90640;  //TODO: not thread safe + size?
		static char bufbind[bufsize];	

		POSITION pos;
		LVITEM item;
		item.mask = LVIF_TEXT;
		item.pszText = bufbind;
		item.cchTextMax = bufsize;
		int subindex(0);
		CString labelseparator(": ");
		labelseparator.Replace(Settings.General.Separator,"");//remove separator, otherwise it will split
		//get item
		pos = loglist.GetFirstSelectedItemPosition();
		if(pos!=NULL){
			item.iItem = loglist.GetNextSelectedItem(pos);
			//get subitems
			while(subindex<ColumnCount){
				item.iSubItem = subindex;
				if(loglist.GetItem(&item)){
					if(subindex<FieldList.GetSize()){ //get fieldlabels if there are any
						Row += FieldList.GetAt(subindex);
						Row += labelseparator;
					}
					Row += item.pszText;
					Row += Settings.General.Separator;
				}
				subindex++;
			}
		}
	}
	return Row;
}

int CLogViewList::EntryCount() const
{
	return loglist.GetItemCount();
}

int CLogViewList::FilteredCount() const
{
	return filtered_count;
}

void CLogViewList::AddFilter(CString& Text, int FilterType)
{
	if(Text.GetLength()>0){
		switch(FilterType){
			case 0:
				Settings.Filter.Include.AddTail(Text);
				break;
			case 1:
				Settings.Filter.Exclude.AddTail(Text);
				break;
		}
		RefreshLog();
	}
}

void CLogViewList::ClearFilters()
{
	Settings.Filter.Include.RemoveAll();
	Settings.Filter.Exclude.RemoveAll();
	RefreshLog();
}

int CLogViewList::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	if (CView::OnCreate(lpCreateStruct) == -1)
		return -1;

	//Add your specialized creation code here

	//set font
	LOGFONT logfont;
	logfont.lfHeight = 14;
	logfont.lfWidth = 0;
	logfont.lfEscapement = 0;
	logfont.lfOrientation = 0;
	logfont.lfWeight = FW_NORMAL;
	logfont.lfItalic = FALSE;
	logfont.lfUnderline = FALSE;
	logfont.lfStrikeOut = FALSE;
	logfont.lfCharSet = ANSI_CHARSET;
	logfont.lfOutPrecision = OUT_DEFAULT_PRECIS;
	logfont.lfClipPrecision = CLIP_DEFAULT_PRECIS;
	logfont.lfQuality = DEFAULT_QUALITY;
	logfont.lfPitchAndFamily = DEFAULT_PITCH;
	strcpy(logfont.lfFaceName,"Verdana");
	font.CreateFontIndirect(&logfont);

	CRect cr;
	GetClientRect(&cr);
	if (!loglist.Create(WS_CHILD|WS_VISIBLE|WS_TABSTOP|WS_BORDER|LVS_REPORT|LVS_SINGLESEL|LVS_SHOWSELALWAYS,cr,this,ID_LIST)){
		return -1;
	}else{
		loglist.SetFont(&font);
		loglist.SetExtendedStyle(LVS_EX_GRIDLINES|LVS_EX_FULLROWSELECT);
	}

	return 0;
}

void CLogViewList::OnSize(UINT nType, int cx, int cy)
{
	CView::OnSize(nType, cx, cy);

	//Add your message handler code here
	CRect cr;
	GetClientRect(&cr);
	loglist.MoveWindow(&cr);
}

BOOL CLogViewList::OnNotify(WPARAM wParam, LPARAM lParam, LRESULT* pResult)
{
	//Add your specialized code here and/or call the base class
	if(wParam==ID_LIST){

		LPNMHDR lpnmhdr;

#if (_WIN32_IE >= 0x0400)
		//version 4.71 ->
		LPNMITEMACTIVATE lpnmia;
		lpnmia = (LPNMITEMACTIVATE) lParam;
		lpnmhdr = (LPNMHDR) (&(lpnmia->hdr));
#else
		//version 0 -> 4.70
	    lpnmhdr = (LPNMHDR) lParam;
#endif

		if(lpnmhdr->code==NM_CLICK){
			ReadLogAppend();	//Refresh
		}
		if(lpnmhdr->code==LVN_KEYDOWN){
			LPNMLVKEYDOWN pnkd = (LPNMLVKEYDOWN) lParam;
			if(pnkd->wVKey == VK_F5){
				ReadLogAppend();	//Refresh
			}
		}
		if(lpnmhdr->code==LVN_ITEMACTIVATE || lpnmhdr->code==LVN_ITEMCHANGED){
			/* if(!AbortBinding){		//if not reading
				ReadLogAppend();	//Refresh
			} //*/
			BindDetailToList();
		}

		//popup menu
		if(lpnmhdr->code==NM_RCLICK){
			
			/*
			TODO: Popup Menu?
			CPoint point;                                            
			::GetCursorPos(&point); //where is the mouse?
			CMenu menu; //lets display our context menu :) 
			DWORD dwSelectionMade;                                       
			VERIFY(menu.LoadMenu(IDR_MENU1) );  
			CMenu *pmenuPopup = menu.GetSubMenu(0);
			ASSERT(pmenuPopup != NULL);                                       
			dwSelectionMade = pmenuPopup->TrackPopupMenu( (TPM_LEFTALIGN|TPM_LEFTBUTTON|
															   TPM_NONOTIFY|TPM_RETURNCMD),
															   point.x, point.y, this
										);                                

			pmenuPopup->DestroyMenu(); 
			//*/
		}
	}
	return CView::OnNotify(wParam, lParam, pResult);
}

/* 
 * SORT
 * http://support.microsoft.com/kb/250614
 * http://msdn.microsoft.com/en-us/library/bb387110.aspx
 * http://www.codeproject.com/KB/list/clistctrl_sort.aspx
 */
void CLogViewList::OnItemclickListSort(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NMLISTVIEW *pLV = (NMLISTVIEW *) pNMHDR;
	bool sortasc(true);

	//remember sort if ascending, next time descending
	if(pLV->iItem==Settings.General.SortedColumn){
		sortasc=false;
		Settings.General.SortedColumn = -1;
	}else{
		Settings.General.SortedColumn = pLV->iItem;
	}

	//sort
	SortColumn(pLV->iItem,sortasc);

	*pResult = 0;
}

void CLogViewList::SortColumn(int ColumnIndex, bool Ascending)
{
    PARAMSORT par(&loglist, ColumnIndex, Ascending);
    loglist.SortItemsEx((PFNLVCOMPARE)SortFunc, (LPARAM)&par);
}

int CALLBACK CLogViewList::SortFunc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	PARAMSORT&		  ps = *(PARAMSORT*)lParamSort;
    CListCtrl* pListCtrl = (CListCtrl*) ps.m_pList;
    CString		strItem1 = pListCtrl->GetItemText(static_cast<int>(lParam1), ps.m_ColumnIndex);
    CString		strItem2 = pListCtrl->GetItemText(static_cast<int>(lParam2), ps.m_ColumnIndex);

	if(ps.m_Ascending){
		return _tcscmp(strItem1,strItem2);
	}else{
		return _tcscmp(strItem2,strItem1);
	}
}



